/**
 * Back 2 Browser Bytecode Translator
 * Copyright (C) 2012-2015 Jaroslav Tulach <jaroslav.tulach@apidesign.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. Look for COPYING file in the top folder.
 * If not, see http://opensource.org/licenses/GPL-2.0.
 */
package org.apidesign.bck2brwsr.flow;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.List;
import javax.script.Invocable;
import javax.script.ScriptContext;
import javax.script.ScriptEngine;
import javax.script.ScriptEngineManager;
import javax.script.ScriptException;
import org.apidesign.vm4brwsr.Bck2Brwsr;
import org.apidesign.vm4brwsr.Bck2Brwsr.Flow;
import static org.testng.Assert.*;

public final class TestVM {
    private final ScriptEngine js;
    private Invocable code;
    private final CharSequence codeSeq;
    private Object bck2brwsr;
    
    
    private TestVM(ScriptEngine js, CharSequence codeSeq) throws ScriptException, NoSuchMethodException {
        this.js = js;
        this.codeSeq = codeSeq;
    }
    
    public Object execCode(
        String msg, Class<?> clazz, String method, 
        Object expRes, Object... args
    ) throws Exception {
        Object ret = null;
        try {
            ret = getCode().invokeMethod(bck2brwsr, "loadClass", clazz.getName());
            List<Object> ma = new ArrayList<Object>();
            ma.add(method);
            ma.addAll(Arrays.asList(args));
            ret = getCode().invokeMethod(ret, "invoke", ma.toArray());
        } catch (ScriptException ex) {
            fail("Execution failed in " + dumpJS(codeSeq) + ": " + ex.getMessage(), ex);
        } catch (NoSuchMethodException ex) {
            fail("Cannot find method in " + dumpJS(codeSeq), ex);
        }
        if (ret == null && expRes == null) {
            return null;
        }
        if (expRes != null && expRes.equals(ret)) {
            return null;
        }
        if (expRes instanceof Number) {
            // in case of Long it is necessary convert it to number
            // since the Long is represented by two numbers in JavaScript
            try {
                final Object toFP = ((ScriptEngine)getCode()).eval("Number.prototype.toFP");
                if (ret instanceof Long) {
                    ret = getCode().invokeMethod(toFP, "call", ret);
                }
                ret = getCode().invokeFunction("Number", ret);
            } catch (ScriptException ex) {
                fail("Conversion to number failed in " + dumpJS(codeSeq) + ": " + ex.getMessage(), ex);
            } catch (NoSuchMethodException ex) {
                fail("Cannot find global Number(x) function in " + dumpJS(codeSeq) + ": " + ex.getMessage(), ex);
            }
        }
        return ret;
    }
    
    void assertExec(
        String msg, Class clazz, String method, Object expRes, Object... args
    ) throws Exception {
        Object ret = execCode(msg, clazz, method, expRes, args);
        if (ret == null) {
            return;
        }
        if (expRes instanceof Integer && ret instanceof Double) {
            expRes = ((Integer)expRes).doubleValue();
        }
        if (expRes != null && expRes.equals(ret)) {
            return;
        }
        assertEquals(ret, expRes, msg + "was: " + ret + "\n" + dumpJS(codeSeq));
    }    

    static TestVM compileClass(String... names) throws ScriptException, IOException {
        return compileClass(null, GraalFlowAnalyzer.getDefault(), names);
    }
    
    static TestVM compileClass(StringBuilder sb, Flow.Analyzer flow, String... names) throws ScriptException, IOException {
        return compileClass(sb, null, flow, names);
    }

    static TestVM compileClass(StringBuilder sb, 
        ScriptEngine[] eng, Flow.Analyzer flow, String... names
    ) throws ScriptException, IOException {
        return compileClass(sb, eng, flow, new EmulationResources(), names);
    }
    static TestVM compileClass(
        StringBuilder sb, 
        ScriptEngine[] eng, 
        Flow.Analyzer flow,
        Bck2Brwsr.Resources resources, 
        String... names
    ) throws ScriptException, IOException {
        if (sb == null) {
            sb = new StringBuilder();
        }
        Bck2Brwsr.newCompiler()
            .resources(resources)
            .addClasses(names)
            .flowAnalyzer(flow)
            .generate(sb);
        ScriptEngineManager sem = new ScriptEngineManager();
        ScriptEngine js = sem.getEngineByExtension("js");
        if (eng != null) {
            eng[0] = js;
        }
        try {
            return new TestVM(js, sb);
        } catch (Exception ex) {
            if (sb.length() > 2000) {
                sb = dumpJS(sb);
            }
            fail("Could not evaluate:" + ex.getClass() + ":" + ex.getMessage() + "\n" + sb, ex);
            return null;
        }
    }

    Object loadClass(String loadClass, String name) throws ScriptException, NoSuchMethodException {
        return getCode().invokeMethod(bck2brwsr, "loadClass", LoopControl.class.getName());
    }
    
    Object invokeMethod(Object obj, String method, Object... params) throws ScriptException, NoSuchMethodException {
        return getCode().invokeMethod(obj, method, params);
    }

    Object invokeFunction(String methodName, Object... args) throws ScriptException, NoSuchMethodException {
        return getCode().invokeFunction(methodName, args);
    }

    static StringBuilder dumpJS(CharSequence sb) throws IOException {
        File f = File.createTempFile("execution", ".js");
        Writer w = new OutputStreamWriter(new FileOutputStream(f), "UTF-8");
        w.append(sb);
        w.close();
        return new StringBuilder(f.getPath());
    }

    @Override
    public String toString() {
        try {
            return dumpJS(codeSeq).toString();
        } catch (IOException ex) {
            return ex.toString();
        }
    }

    final CharSequence codeSeq() {
        return codeSeq;
    }

    /**
     * @return the code
     */
    private Invocable getCode() throws ScriptException {
        if (code == null) {
            js.eval(codeSeq.toString());
            js.getContext().setAttribute("loader", this, ScriptContext.ENGINE_SCOPE);
            this.bck2brwsr = js.eval("bck2brwsr(function(n) { return loader.get(n); })");
            code = (Invocable) js;
        }
        return (Invocable)code;
    }
    
    private static class EmulationResources implements Bck2Brwsr.Resources {
        @Override
        public InputStream get(String name) throws IOException {
            if ("java/net/URI.class".equals(name)) {
                // skip
                return null;
            }
            if ("java/net/URLConnection.class".equals(name)) {
                // skip
                return null;
            }
            if ("java/lang/System.class".equals(name)) {
                // skip
                return null;
            }
            if ("java/io/PrintStream.class".equals(name)) {
                // skip
                return null;
            }
            if ("java/io/PrintWriter.class".equals(name)) {
                // skip
                return null;
            }
            Enumeration<URL> en = LoopControlTest.class.getClassLoader().getResources(name);
            URL u = null;
            while (en.hasMoreElements()) {
                u = en.nextElement();
            }
            if (u == null) {
                throw new IOException("Can't find " + name);
            }
            if (u.toExternalForm().contains("rt.jar!")) {
                throw new IOException("No emulation for " + u);
            }
            return u.openStream();
        }
    }
}
